from typing import Literal as _Literal, Union as _Union


UnknownErrcode = _Literal["unknown"]

LiqpayAntiFraudErrcode = _Literal["limit", "frod", "decline"]

LiqpayRequestErrcode = _Literal[
    "err_action",
    "invalid_signature",
    "public_key_not_found",
    "order_id_empty",
    "amount_limit",
    "wrong_amount_currency",
]

LiqpayExpireErrcode = _Literal[
    "expired_phone",
    "expired_3ds",
    "expired_otp",
    "expired_cvv",
    "expired_p24",
    "expired_sender",
    "expired_pin",
    "expired_ivr",
    "expired_captcha",
    "expired_password",
    "expired_senderapp",
    "expired_prepared",
    "expired_mp",
    "expired_qr",
]

LiqpayPaymentErrcode = _Literal[
    "payment_err_sender_card",
    "payment_processing",
    "payment_err_status",
    "payment_not_found",
    "payment_not_subscribed",
    "payment_err_type",
]

LiqpayNonFinancialErrcode = _Union[
    _Literal[
        "err_auth",
        "err_cache",
        "err_sms_send",
        "err_sms_otp",
        "err_shop_not_agent",
        "err_card_def_notfound",
        "err_no_card_token",
        "err_card_liqpay_def",
        "err_card_type",
        "err_card_country",
        "err_limit_amount",
        "err_payment_amount_limit",
        "err_payment_discount",
        "err_wallet",
        "err_get_verify_code",
        "err_verify_code",
        "err_path",
        "err_payment_cash_acq",
        "err_split_amount",
        "err_card_receiver_def",
        "err_amount_hold",
        "err_access",
        "err_blocked",
        "err_empty",
        "err_empty_phone",
        "err_missing",
        "err_wrong",
        "err_wrong_currency",
        "err_phone",
        "err_card",
        "err_card_bin",
        "err_terminal_notfound",
        "err_commission_notfound",
        "err_payment_create",
        "err_mpi",
        "err_currency_is_not_allowed",
        "err_look",
        "err_mods_empty",
        "err_payment_currency",
        "err_payment_exchangerates",
        "err_signature",
        "err_api_action",
        "err_api_callback",
        "err_api_ip",
        "user_not_found",
        "wait_info",
        "order_id_duplicate",
        "shop_blocked",
        "shop_not_active",
        "5",
    ],
    LiqpayExpireErrcode,
    LiqpayRequestErrcode,
    LiqpayPaymentErrcode,
]

LiqpayFinancialErrcode = _Literal[
    "90",
    "101",
    "102",
    "103",
    "104",
    "105",
    "106",
    "107",
    "108",
    "109",
    "110",
    "111",
    "112",
    "113",
    "114",
    "115",
    "2903",
    "2915",
    "3914",
    "9851",
    "9852",
    "9854",
    "9855",
    "9857",
    "9859",
    "9860",
    "9861",
    "9863",
    "9867",
    "9868",
    "9872",
    "9882",
    "9886",
    "9961",
    "9989",
]

LiqPayErrcode = _Union[
    UnknownErrcode,
    LiqpayAntiFraudErrcode,
    LiqpayFinancialErrcode,
    LiqpayNonFinancialErrcode,
]



from typing import Literal, TypedDict, Required, List
from numbers import Number
from datetime import datetime, timedelta
from uuid import UUID

from liqpy.models.request import FiscalItem, DetailAddenda, SplitRule, FiscalInfo

from .common import SubscribePeriodicity, Language, PayType, PayOption, Format
from .action import Action


class DetailAddendaDict(TypedDict):
    airLine: str
    ticketNumber: str
    passengerName: str
    flightNumber: str
    originCity: str
    destinationCity: str
    departureDate: str


class SplitRuleDict(TypedDict, total=False):
    public_key: Required[str]
    amount: Required[Number]
    commission_payer: Required[Literal["sender", "receiver"]]
    server_url: Required[str]


class FiscalItemDict(TypedDict):
    id: int
    amount: Number
    cost: Number
    price: Number


class FiscalInfoDict(TypedDict, total=False):
    items: List[FiscalItemDict | FiscalItem]
    delivery_emails: List[str]


class ProductDict(TypedDict, total=False):
    product_category: str
    product_description: str
    product_name: str
    product_url: str


class OneClickDict(TypedDict, total=False):
    customer: str
    recurringbytoken: Literal["1", True]
    customer_user_id: str


class SubscribeDict(TypedDict, total=False):
    subscribe: Literal[1, True]
    subscribe_date_start: str | datetime | timedelta | None
    subscribe_periodicity: SubscribePeriodicity


class LetterDict(TypedDict, total=False):
    letter_of_credit: Literal[1, True]
    letter_of_credit_date: str | datetime | timedelta


class MPIParamsDict(TypedDict, total=False):
    mpi_md: str
    mpi_pares: str


class SenderDict(TypedDict, total=False):
    phone: str
    sender_phone: str
    sender_first_name: str
    sender_last_name: str
    sender_email: str
    sender_country_code: str
    sender_city: str
    sender_address: str
    sender_postal_code: str
    sender_shipping_state: str


class CardDict(TypedDict, total=False):
    card: Required[str]
    card_exp_month: str
    card_exp_year: str
    card_cvv: str


class BaseRequestDict(TypedDict, total=False):
    version: Required[int]
    public_key: Required[str]
    action: Required["Action"]


class LiqpayRequestDict(
    CardDict,
    SenderDict,
    MPIParamsDict,
    LetterDict,
    SubscribeDict,
    OneClickDict,
    ProductDict,
    total=False,
):
    opid: str | int
    order_id: str | UUID
    payment_id: int
    amount: Number | str
    rro_info: FiscalInfoDict | FiscalInfo
    expired_date: str | datetime | timedelta
    language: Language
    paytype: PayType
    paytypes: PayOption
    result_url: str
    server_url: str
    verifycode: Literal["Y", True]
    email: str
    date_from: int | datetime | timedelta
    date_to: int | datetime | timedelta
    resp_format: Format
    split_rules: list[SplitRuleDict | SplitRule]
    dae: DetailAddendaDict | DetailAddenda
    info: str



from pytest import fixture
from typing import TYPE_CHECKING
from datetime import datetime, UTC
from ipaddress import IPv4Address

from liqpy.api import Decoder

from tests import EXAMPLES_DIR

if TYPE_CHECKING:
    from liqpy.types import LiqpayCallbackDict


@fixture
def decoder():
    return Decoder()


def test_decode_status_example(decoder: Decoder):
    with open(EXAMPLES_DIR / "status.json", "r") as fp:
        o: "LiqpayCallbackDict" = decoder.decode(fp.read())

    assert o["create_date"] == datetime(2017, 8, 3, 10, 55, 16, 373000, tzinfo=UTC)
    assert o["end_date"] == datetime(2017, 8, 3, 10, 55, 29, 972000, tzinfo=UTC)
    assert o["ip"] == IPv4Address("8.8.8.8")

    assert isinstance(o["version"], int)
    assert isinstance(o["payment_id"], int)
    assert isinstance(o["transaction_id"], int)
    assert isinstance(o["acq_id"], int)
    assert isinstance(o["amount"], float)
    assert isinstance(o["mpi_eci"], int)



from decimal import Decimal
from typing import Self
from enum import StrEnum, auto, Enum, member
from datetime import datetime
from dataclasses import dataclass

from liqpy.util.convert import to_datetime


class Field(Enum):
    ID = member(lambda x: int(x))
    AMOUNT = member(lambda x: Decimal(x))
    SENDER_COMMISSION = member(lambda x: Decimal(x))
    RECEIVER_COMMISSION = member(lambda x: Decimal(x))
    AGENT_COMMISSION = member(lambda x: Decimal(x))
    CURRENCY = member(lambda x: Currency(x))
    AMOUNT_CREDIT = member(lambda x: Decimal(x) if x else None)
    COMISSION_CREDIT = member(lambda x: Decimal(x) if x else None)
    CURRENCY_CREDIT = member(lambda x: Currency(x) if x else None)
    CREATE_DATE = member(lambda x: to_datetime(x))
    END_DATE = member(lambda x: to_datetime(x))
    TYPE = member(lambda x: Type(x))
    STATUS = member(lambda x: Status(x))
    STATUS_ERR_CODE = member(lambda x: Code(x) if x else None)
    AUTH_CODE = member(lambda x: x if x else None)
    SHOP_ORDER_ID = member(lambda x: x)
    DESCRIPTION = member(lambda x: x)
    PHONE = member(lambda x: x if x else None)
    SENDER_COUNTRY_CODE = member(lambda x: x if x else None)
    CARD = member(lambda x: x if x else None)
    ISSUER_BANK = member(lambda x: x if x else None)
    CARD_COUNTRY = member(lambda x: x if x else None)
    CARD_TYPE = member(lambda x: x if x else None)
    PAY_WAY = member(lambda x: PayWay(x) if x else None)
    RECEIVER_CARD = member(lambda x: x if x else None)
    RECEIVER_OKPO = member(lambda x: int(x) if x else None)
    REFUND_AMOUNT = member(lambda x: Decimal(x) if x else None)
    REFUND_DATE_LAST = member(lambda x: to_datetime(x) if x else None)
    REFUND_RESERVE_IDS = member(
        lambda x: frozenset(map(lambda v: v.strip(), x.split("|")) if x else [])
    )
    RESERVE_REFUND_ID = member(lambda x: int(x) if x else None)
    RESERVE_PAYMENT_ID = member(lambda x: int(x) if x else None)
    RESERVE_AMOUNT = member(lambda x: Decimal(x) if x else None)
    RESERVE_DATE = member(lambda x: to_datetime(x) if x else None)
    COMPLETION_DATE = member(lambda x: to_datetime(x) if x else None)
    INFO = member(lambda x: x)
    LIQPAY_ORDER_ID = member(lambda x: x if x else None)
    COMPENSATION_ID = member(lambda x: x if x else None)
    COMPENSATION_DATE = member(lambda x: to_datetime(x) if x else None)
    BONUSPLUS_ACCOUNT = member(lambda x: x if x else None)
    BONUS_TYPE = member(lambda x: x if x else None)
    BONUS_PERCENT = member(lambda x: Decimal(x) if x else None)
    BONUS_AMOUNT = member(lambda x: Decimal(x) if x else None)


class Currency(StrEnum):
    UAH = "UAH"
    USD = "USD"
    EUR = "EUR"


class Action(StrEnum):
    PAY = auto()
    HOLD = auto()
    PAYSPLIT = auto()
    SUBSCRIBE = auto()
    PAYDONATE = auto()
    AUTH = auto()
    REGULAR = auto()


class Status(StrEnum):
    # final payment statuses
    SUCCESS = auto()
    ERROR = auto()
    FAILURE = auto()
    REVERSED = auto()

    # subscription statuses
    SUBSCRIBED = auto()
    UNSUBSCRIBED = auto()

    # statuses that required payment confirmation
    VERIFY_3DS = "3ds_verify"
    VERIFY_CVV = "cvv_verify"
    VERIFY_OTP = "otp_verify"
    VERIFY_RECEIVER = "receiver_verify"
    VERIFY_SENDER = "sender_verify"

    # other payment statuses
    WAIT_ACCEPT = auto()
    WAIT_SECURE = auto()

    def is_wait(self) -> bool:
        return self in (Status.WAIT_ACCEPT, Status.WAIT_SECURE)

    def requires_confirmation(self) -> bool:
        return self in (
            Status.VERIFY_3DS,
            Status.VERIFY_CVV,
            Status.VERIFY_OTP,
            Status.VERIFY_RECEIVER,
            Status.VERIFY_SENDER,
        )

    def is_final(self) -> bool:
        return self in (
            Status.SUCCESS,
            Status.ERROR,
            Status.FAILURE,
            Status.REVERSED,
        )


class PayWay(StrEnum):
    CARD = auto()
    LIQPAY = auto()
    PRIVAT24 = auto()
    MASTERPASS = auto()
    MOMENT_PART = auto()
    CASH = auto()
    INVOICE = auto()
    QR = auto()

    # Google Pay
    GPAY = auto()

    # Google Pay Card
    GPAYCARD = auto()

    # Apple Pay
    APAY = auto()


class Code(StrEnum):
    ERR_BLOCKED = auto()
    ERR_CARD_BIN = auto()
    ERR_TOKEN_DECODE = auto()
    ERR_MISSING = auto()

    LIMIT = auto()

    # expired codes
    EXPIRED = auto()
    EXPIRED_P24 = auto()
    EXPIRED_3DS = auto()

    MPI = auto()

    E4 = "4"

    # financial errors
    E90 = "90"
    E101 = "101"
    E102 = "102"
    E103 = "103"
    E104 = "104"
    E105 = "105"
    E106 = "106"
    E107 = "107"
    E108 = "108"
    E109 = "109"
    E110 = "110"
    E111 = "111"
    E112 = "112"
    E113 = "113"
    E114 = "114"
    E115 = "115"
    E2903 = "2903"
    E2915 = "2915"
    E3914 = "3914"
    E9851 = "9851"
    E9852 = "9852"
    E9854 = "9854"
    E9855 = "9855"
    E9857 = "9857"
    E9859 = "9859"
    E9860 = "9860"
    E9861 = "9861"
    E9863 = "9863"
    E9867 = "9867"
    E9868 = "9868"
    E9872 = "9872"
    E9875 = "9875"
    E9882 = "9882"
    E9886 = "9886"
    E9961 = "9961"
    E9989 = "9989"


class Type(StrEnum):
    ACQUIRING = auto()


@dataclass(frozen=True, eq=False, slots=True)
class Report:
    """
    Dataclass model for a LiqPay report comma-separated entry (CSV)
    """

    id: int
    shop_order_id: str
    liqpay_order_id: str

    amount: Decimal
    currency: Currency

    sender_commission: Decimal
    receiver_commission: Decimal
    agent_commission: Decimal

    create_date: datetime
    end_date: datetime

    type: Type

    status: Status

    description: str

    receiver_card: str | None = None
    receiver_okpo: int | None = None

    card: str | None = None
    issuer_bank: str | None = None
    card_country: str | None = None
    card_type: str | None = None

    info: str = ""

    phone: str | None = None

    sender_country_code: str | None = None

    pay_way: PayWay | None = None

    amount_credit: Decimal | None = None
    comission_credit: Decimal | None = None
    currency_credit: Currency | None = None

    auth_code: str | None = None

    status_err_code: Code | None = None

    refund_amount: Decimal | None = None
    refund_date_last: datetime | None = None
    refund_reserve_ids: frozenset[str] = frozenset()

    reserve_refund_id: int | None = None
    reserve_payment_id: int | None = None
    reserve_amount: Decimal | None = None
    reserve_date: datetime | None = None

    completion_date: datetime | None = None

    compensation_id: str | None = None
    compensation_date: datetime | None = None

    bonusplus_account: str | None = None
    bonus_type: str | None = None
    bonus_percent: Decimal | None = None
    bonus_amount: Decimal | None = None

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> "Report":
        """
        Create a new instance from a dictionary of strings

        Useful
        """
        return cls(**{k.lower(): Field[k].value(v) for k, v in data.items()})



from typing import Literal as _Literal, Union as _Union


SubscriptionStatus = _Literal["subscribed", "unsubscribed"]
ErrorStatus = _Literal["error", "failure"]
SuccessStatus = _Literal["success"]
FinalStatus = _Union[ErrorStatus, _Literal["reversed"], SuccessStatus]
ConfirmationStatus = _Literal[
    "3ds_verify",
    "captcha_verify",
    "cvv_verify",
    "ivr_verify",
    "otp_verify",
    "password_verify",
    "phone_verify",
    "pin_verify",
    "receiver_verify",
    "sender_verify",
    "senderapp_verify",
    "wait_qr",
    "wait_sender",
]
OtherStatus = _Literal[
    "cash_wait",
    "hold_wait",
    "invoice_wait",
    "prepared",
    "processing",
    "wait_accept",
    "wait_card",
    "wait_compensation",
    "wait_lc",
    "wait_reserve",
    "wait_secure",
]
Status = CallbackStatus = _Union[
    SubscriptionStatus, FinalStatus, ConfirmationStatus, OtherStatus
]



from json import JSONDecoder
from ipaddress import IPv4Address

from liqpy.util.convert import from_milliseconds


class Decoder(JSONDecoder):
    """Custom JSON decoder for LiqPay API responses"""

    def __init__(self):
        super().__init__(
            object_hook=self._object_hook,
            parse_float=float,
            parse_int=int,
            parse_constant=None,
            strict=True,
            object_pairs_hook=None,
        )
        self.mpi_eci = int
        self.ip = IPv4Address

        self.create_date = from_milliseconds
        self.end_date = from_milliseconds
        self.reserve_date = from_milliseconds
        self.completion_date = from_milliseconds
        self.refund_date_last = from_milliseconds
    
    def _object_hook(self, o: dict, /) -> dict:
        for key, value in o.items():
            try:
                fn = getattr(self, key, None)

                if not callable(fn):
                    continue

                processed = fn(value)

                if processed is not None:
                    o[key] = processed

            except Exception as e:
                raise Exception(f"Failed to post convert {key} parameter.") from e

        return o
    


from typing import Literal
from dataclasses import dataclass
from datetime import date
from numbers import Number

from liqpy.util.convert import to_date


@dataclass(kw_only=True)
class DetailAddenda:
    air_line: str
    ticket_number: str
    passenger_name: str
    flight_number: str
    origin_city: str
    destination_city: str
    departure_date: date

    def __post_init__(self):
        self.departure_date = to_date(self.departure_date)

    @classmethod
    def from_json(cls, data: dict):
        s: str = data.get("departureDate") or data.get("departure_date")
        return cls(
            air_line=data.get("airLine") or data.get("air_line"),
            ticket_number=data.get("ticketNumber") or data.get("ticket_number"),
            passenger_name=data.get("passengerName") or data.get("passenger_name"),
            flight_number=data.get("flightNumber") or data.get("flight_number"),
            origin_city=data.get("originCity") or data.get("origin_city"),
            destination_city=data.get("destinationCity")
            or data.get("destination_city"),
            departure_date=date(2000 + int(s[:2]), int(s[2:4]), int(s[4:])),
        )

    def to_json(self):
        return {
            "airLine": self.air_line,
            "ticketNumber": self.ticket_number,
            "passengerName": self.passenger_name,
            "flightNumber": self.flight_number,
            "originCity": self.origin_city,
            "destinationCity": self.destination_city,
            "departureDate": self.departure_date.strftime(r"%y%m%d"),
        }


@dataclass(kw_only=True)
class SplitRule:
    public_key: str
    amount: Number
    commission_payer: Literal["sender", "receiver"]
    server_url: str


@dataclass(kw_only=True)
class FiscalItem:
    id: int
    amount: Number
    cost: Number
    price: Number


@dataclass(kw_only=True)
class FiscalInfo:
    items: list[FiscalItem]
    delivery_emails: list[str]
